
# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-03-01T01:32:59-0500",
 "dirty": false,
 "error": null,
 "full-revisionid": "8f20334f0f5208f4f4b38a6998aadf821ae8bff6",
 "version": "v1.3.2"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
